#ifndef TILE_H
#define TILE_H

class Tile
{
public: 
    Uint8 priority;
    Uint8 paletteLine;
    Uint8 xFlip;
    Uint8 yFlip;
    Uint16 tileID;
public:
    Tile();
    void ReadTile(FILE* file);
    void WriteTile(FILE* file);
    void ClearTile();
    void SetPal(Uint8 paletteLine);
    void FlipX() {this->xFlip ^= 0x08;}
    void FlipY() {this->yFlip ^= 0x10;}
    void SwapPriority() {this->priority ^= 0x80;}
    void SetID(int ID);
};

Tile::Tile() {
    this->tileID = 0;
    this->paletteLine = 0;
    this->xFlip = 0;
    this->yFlip = 0;
    this->priority = 0;
}

void Tile::ReadTile(FILE* file) {
    Uint8 byte = fgetc(file);
    this->tileID = fgetc(file);
    this->paletteLine = (byte & 0x60) >> 5;
    this->xFlip = byte & 0x08;
    this->yFlip = byte & 0x10;
    this->priority = byte & 0x80;
    this->tileID += (byte & 0x07)<<8;
}

void Tile::WriteTile(FILE* file) {
    fputc((char) (this->priority
        | this->xFlip
        | this->yFlip
        | (this->paletteLine << 5)
        | (this->tileID >> 8)), file);
    fputc(this->tileID & 0xFF, file);
}

void Tile::ClearTile() {
    this->tileID = 0;
    this->paletteLine = 0;
    this->xFlip = 0;
    this->yFlip = 0;
    this->priority = 0;
}

void Tile::SetPal(Uint8 paletteLine) {
    this->paletteLine = paletteLine;
}

void Tile::SetID(int ID) {
    this->tileID = ID;
    this->xFlip = 0;
    this->yFlip = 0;
    this->priority = 0;
}

#endif
